////////////////////////////////////////////////////////////////////////////
//                            **** NOCLIP ****                            //
//                  Copyright (c) 2007 Conifer Software                   //
//                          All Rights Reserved.                          //
//      Distributed under the BSD Software License (see license.txt)      //
////////////////////////////////////////////////////////////////////////////

// noclip.c

#if defined(_WIN32)
#include <windows.h>
#include <io.h>
typedef unsigned __int32 uint32_t;
typedef __int32 int32_t;
#else
#include <inttypes.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>
#include <stdio.h>

#ifndef PATH_MAX
#ifdef MAX_PATH
#define PATH_MAX MAX_PATH
#elif defined (MAXPATHLEN)
#define PATH_MAX MAXPATHLEN
#else
#define PATH_MAX 1024
#endif
#endif

// RIFF / wav header formats

typedef struct {
    char ckID [4];
    uint32_t ckSize;
    char formType [4];
} RiffChunkHeader;

typedef struct {
    char ckID [4];
    uint32_t ckSize;
} ChunkHeader;

typedef struct {
    unsigned short FormatTag, NumChannels;
    uint32_t SampleRate, BytesPerSecond;
    unsigned short BlockAlign, BitsPerSample;
    unsigned short cbSize, ValidBitsPerSample;
    int32_t ChannelMask;
    unsigned short SubFormat;
    char GUID [14];
} WaveHeader;

///////////////////////////// local variable storage //////////////////////////

static const char *sign_on = "\n"
" NOCLIP  Clipping Prevention Preprocessor  Version 0.1\n"
" Copyright (c) 2007 Conifer Software.  All Rights Reserved.\n\n";

static const char *usage =
" Usage:   NOCLIP [-options] [infile.wav|-] [outfile.wav|-]\n\n"
" Options: -pn = set peak output level, n = 0.900 to 0.999 (default = 0.980)\n"
"          -q  = quiet (keep console output to a minimum)\n"
"          -s  = display summary statistics when done processing\n"
"          -y  = overwrite existing file (use with caution!)\n\n";

int quiet_mode, overwrite, statistics;

/////////////////////////// local function declarations ////////////////////////

static int process_file (char *infilename, char *outfilename, double peak_limit);
static int process_audio (FILE *infile, FILE *outfile, int num_samples, int num_channels, int sample_rate, double peak_limit);
static int DoReadFile (FILE *hFile, void *lpBuffer, uint32_t nNumberOfBytesToRead, uint32_t *lpNumberOfBytesRead);
static int DoWriteFile (FILE *hFile, void *lpBuffer, uint32_t nNumberOfBytesToWrite, uint32_t *lpNumberOfBytesWritten);
static double softclip (int input, int max_in, int max_out, int knee);
static void string_time (char *time_string, double seconds);

#define NO_ERROR 0L
#define SOFT_ERROR 1
#define HARD_ERROR 2

#define MAX_CHANNELS	8
#define ATTACK_TC	0.05
#define DECAY_TC	0.5

//////////////////////////////////// main //////////////////////////////////////

int main (argc, argv) int argc; char **argv;
{
    char *infilename = NULL, *outfilename = NULL;
    double peak_limit = 0.98;
    int error_count = 0;

    // loop through command-line arguments

    while (--argc)
#if defined (WIN32)
        if ((**++argv == '-' || **argv == '/') && (*argv)[1])
#else
        if ((**++argv == '-') && (*argv)[1])
#endif
            while (*++*argv)
                switch (**argv) {

                    case 'Y': case 'y':
                        overwrite = 1;
                        break;

                    case 'P': case 'p':
                        peak_limit = strtod (++*argv, argv);
                        --*argv;

                        if (peak_limit < 0.900 || peak_limit > 0.999) {
                            fprintf (stderr, "peak value spec must be 0.900 to 0.999!\n");
                            ++error_count;
                        }

                        break;

                    case 'S': case 's':
                        statistics = 1;
                        break;

                    case 'Q': case 'q':
                        quiet_mode = 1;
                        break;

                    default:
                        fprintf (stderr, "illegal option: %c !\n", **argv);
                        ++error_count;
                }
        else if (!infilename) {
            infilename = malloc (strlen (*argv) + PATH_MAX);
            strcpy (infilename, *argv);
        }
        else if (!outfilename) {
            outfilename = malloc (strlen (*argv) + PATH_MAX);
            strcpy (outfilename, *argv);
        }
        else {
            fprintf (stderr, "extra unknown argument: %s !\n", *argv);
            ++error_count;
        }

    // check for various command-line argument problems

    if (!quiet_mode && !error_count)
        fprintf (stderr, sign_on);

    if (!infilename || !outfilename || error_count) {
        printf ("%s", usage);
        return 1;
    }

    return process_file (infilename, outfilename, peak_limit);
}

static int process_file (char *infilename, char *outfilename, double peak_limit)
{
    uint32_t total_samples = 0, bcount;
    RiffChunkHeader riff_chunk_header;
    ChunkHeader chunk_header;
    WaveHeader WaveHeader;
    FILE *infile, *outfile;
    int result;

    // open the source file for reading

    if (*infilename == '-') {
        infile = stdin;
#if defined(WIN32)
        _setmode (fileno (stdin), O_BINARY);
#endif
    }
    else if ((infile = fopen (infilename, "rb")) == NULL) {
        fprintf (stderr, "can't open file %s!\n", infilename);
        return SOFT_ERROR;
    }

    // check output file for overwrite warning required

    if (*outfilename != '-' && !overwrite && (outfile = fopen (outfilename, "rb")) != NULL) {
        fprintf (stderr, "file already exists, use -y to override\n");
        fclose (outfile);
        fclose (infile);
        return SOFT_ERROR;
    }

    // open output file for writing

    if (*outfilename == '-') {
        outfile = stdout;
#if defined(WIN32)
        _setmode (fileno (stdout), O_BINARY);
#endif
    }
    else if ((outfile = fopen (outfilename, "w+b")) == NULL) {
        fprintf (stderr, "can't create file %s!\n", outfilename);
        fclose (infile);
        return SOFT_ERROR;
    }

    // read (and copy to output) initial RIFF form header

    if (1) {
        if ((!DoReadFile (infile, &riff_chunk_header, sizeof (RiffChunkHeader), &bcount) ||
            bcount != sizeof (RiffChunkHeader) || strncmp (riff_chunk_header.ckID, "RIFF", 4) ||
            strncmp (riff_chunk_header.formType, "WAVE", 4))) {
                fprintf (stderr, "%s is not a valid .WAV file!\n", infilename);
                fclose (infile);
                fclose (outfile);
                remove (outfilename);
                return SOFT_ERROR;
        }
        else if (!DoWriteFile (outfile, &riff_chunk_header, sizeof (RiffChunkHeader), &bcount) ||
                bcount != sizeof (RiffChunkHeader)) {
                    fprintf (stderr, "can't write data, disk probably full!\n");
                    fclose (infile);
                    fclose (outfile);
                    remove (outfilename);
                    return SOFT_ERROR;
        }
    }

    // loop through all elements of the RIFF wav header
    // (until the data chuck) and copy them to the output file

    while (1) {
        if (!DoReadFile (infile, &chunk_header, sizeof (ChunkHeader), &bcount) ||
            bcount != sizeof (ChunkHeader)) {
                fprintf (stderr, "%s is not a valid .WAV file!\n", infilename);
                fclose (infile);
                fclose (outfile);
                remove (outfilename);
                return SOFT_ERROR;
        }
        else if (!DoWriteFile (outfile, &chunk_header, sizeof (ChunkHeader), &bcount) ||
            bcount != sizeof (ChunkHeader)) {
                fprintf (stderr, "can't write data, disk probably full!\n");
                fclose (infile);
                fclose (outfile);
                remove (outfilename);
                return SOFT_ERROR;
        }

        // if it's the format chunk, we want to get some info out of there and
        // make sure it's a .wav file we can handle

        if (!strncmp (chunk_header.ckID, "fmt ", 4)) {
            int supported = 1, bits_per_sample, format;

            if (chunk_header.ckSize < 16 || chunk_header.ckSize > sizeof (WaveHeader) ||
                !DoReadFile (infile, &WaveHeader, chunk_header.ckSize, &bcount) ||
                bcount != chunk_header.ckSize) {
                    fprintf (stderr, "%s is not a valid .WAV file!\n", infilename);
                    fclose (infile);
                    fclose (outfile);
                    remove (outfilename);
                    return SOFT_ERROR;
            }
            else if (!DoWriteFile (outfile, &WaveHeader, chunk_header.ckSize, &bcount) ||
                bcount != chunk_header.ckSize) {
                    fprintf (stderr, "can't write data, disk probably full!\n");
                    fclose (infile);
                    fclose (outfile);
                    remove (outfilename);
                    return SOFT_ERROR;
            }

            format = (WaveHeader.FormatTag == 0xfffe && chunk_header.ckSize == 40) ?
                WaveHeader.SubFormat : WaveHeader.FormatTag;

            bits_per_sample = chunk_header.ckSize == 40 ?
                WaveHeader.ValidBitsPerSample : WaveHeader.BitsPerSample;

            if (format != 1)
                supported = 0;

            if (!WaveHeader.NumChannels ||
                WaveHeader.BlockAlign / WaveHeader.NumChannels < (bits_per_sample + 7) / 8 ||
                WaveHeader.BlockAlign / WaveHeader.NumChannels > 4 ||
                WaveHeader.BlockAlign % WaveHeader.NumChannels)
                    supported = 0;

            if (bits_per_sample != 16 || WaveHeader.NumChannels > MAX_CHANNELS)
                supported = 0;

            if (!supported) {
                fprintf (stderr, "%s is an unsupported .WAV format!\n", infilename);
                fclose (infile);
                fclose (outfile);
                remove (outfilename);
                return SOFT_ERROR;
            }
        }
        else if (!strncmp (chunk_header.ckID, "data", 4)) {

            // on the data chunk, get size and exit loop

            total_samples = chunk_header.ckSize / WaveHeader.BlockAlign;

            if (!total_samples) {
                fprintf (stderr, "this .WAV file has no audio samples, probably is corrupt!\n");
                fclose (infile);
                fclose (outfile);
                remove (outfilename);
                return SOFT_ERROR;
            }

            break;
        }
        else {          // just copy unknown chunks to output file

            int bytes_to_copy = (chunk_header.ckSize + 1) & ~1L;
            char *buff = malloc (bytes_to_copy);

            if (!DoReadFile (infile, buff, bytes_to_copy, &bcount) || bcount != bytes_to_copy ||
                !DoWriteFile (outfile, buff, bytes_to_copy, &bcount) || bcount != bytes_to_copy) {
                    fprintf (stderr, "can't write data, disk probably full!\n");
                    fclose (infile);
                    fclose (outfile);
                    remove (outfilename);
                    free (buff);
                    return SOFT_ERROR;
            }

            free (buff);
        }
    }

    result = process_audio (infile, outfile, total_samples, WaveHeader.NumChannels,
	WaveHeader.SampleRate, peak_limit);

    // if everything went well try to read anything else that might be appended to
    // the audio data and write that to the output file

    if (result == NO_ERROR) {
        unsigned char buff [16];
        uint32_t bcount2;

        while (DoReadFile (infile, buff, sizeof (buff), &bcount) && bcount)
            if (!DoWriteFile (outfile, buff, bcount, &bcount2) || bcount2 != bcount) {
                fprintf (stderr, "can't write data, disk probably full!\n");
                result = HARD_ERROR;
                break;
            }
    }

    fclose (infile);     // we're now done with input file, so close

    if (fclose (outfile)) {
        fprintf (stderr, "can't close file!\n");

        if (result == NO_ERROR)
            result = SOFT_ERROR;
    }

    if (result != NO_ERROR) {
        remove (outfilename);
        return result;
    }

    return NO_ERROR;
}

#define INPUT_SAMPLES 65536
#define MAX_SAMPLE 32768

struct chan_data {
    double rms_attack, rms_decay, level_max, level_sum;
    int loud_samples, trans_samples, soft_samples;
    int in_max, in_min, out_max, out_min;
};

static int process_audio (FILE *infile, FILE *outfile, int num_samples, int num_channels,
    int sample_rate, double peak_limit)
{
    int samples_remaining = num_samples, samples_done = 0;
    int trans_begin, trans_end, soft_knee, max_out_value;
    struct chan_data chan_data [MAX_CHANNELS];
    double attack_factor, decay_factor;
    short *sample_buffer;
    int progress = -1.0;

    memset (chan_data, 0, sizeof (chan_data));

    max_out_value = (int)(MAX_SAMPLE * peak_limit);
    soft_knee = (int)(MAX_SAMPLE * (1.0 - (1.0 - peak_limit) * 5.0));
    trans_end = soft_knee / 4;
    trans_begin = soft_knee / 16;

    attack_factor = 1.0 - (2.0 / ATTACK_TC / sample_rate);
    decay_factor = 1.0 - (2.0 / DECAY_TC / sample_rate);

    if (!quiet_mode)
	fprintf (stderr, "processing,");

    sample_buffer = malloc (INPUT_SAMPLES * num_channels * 2);

    while (1) {
        uint32_t bytes_to_read, bytes_read = 0, bcount;
        int sample_count;

        if (samples_remaining > INPUT_SAMPLES)
            bytes_to_read = INPUT_SAMPLES * num_channels * 2;
        else
            bytes_to_read = samples_remaining * num_channels * 2;

        DoReadFile (infile, sample_buffer, bytes_to_read, &bytes_read);
        sample_count = bytes_read / num_channels / 2;
        samples_remaining -= sample_count;

        if (!sample_count)
            break;

        if (sample_count) {
            short *sptr = sample_buffer;
            int c, sc = sample_count;

            while (sc--) {
                for (c = 0; c < num_channels; ++c) {
                    double value = sptr [c], level;

                    if (sptr [c] > chan_data [c].in_max)
                        chan_data [c].in_max = sptr [c];

                    if (sptr [c] < chan_data [c].in_min)
                        chan_data [c].in_min = sptr [c];

                    chan_data [c].rms_attack = chan_data [c].rms_attack * attack_factor +
                        ((double) value * value) * (1.0 - attack_factor);

                    if (chan_data [c].rms_attack > chan_data [c].rms_decay)
                        chan_data [c].rms_decay = chan_data [c].rms_attack;
                    else
                        chan_data [c].rms_decay *= decay_factor;

                    level = sqrt (chan_data [c].rms_decay);

                    if (level > chan_data [c].level_max)
                        chan_data [c].level_max = level;

		    chan_data [c].level_sum += level;

		    if (level > trans_end) {
			chan_data [c].loud_samples++;
			value *= peak_limit;
		    }
		    else if (level > trans_begin) {
			double loud = value * peak_limit;
			double soft = softclip (sptr [c], MAX_SAMPLE, max_out_value, soft_knee);
			double ratio = (level - trans_begin) / (trans_end - trans_begin);

			value = (ratio * loud) + ((1.0 - ratio) * soft);
			chan_data [c].trans_samples++;
		    }
		    else {
			value = softclip (sptr [c], MAX_SAMPLE, max_out_value, soft_knee);
			chan_data [c].soft_samples++;
		    }

		    sptr [c] = (int) floor (value + 0.5);

                    if (sptr [c] > chan_data [c].out_max)
                        chan_data [c].out_max = sptr [c];

                    if (sptr [c] < chan_data [c].out_min)
                        chan_data [c].out_min = sptr [c];
                }

                sptr += num_channels;
            }
        }

        if (sample_count && !DoWriteFile (outfile, sample_buffer, sample_count * num_channels * 2, &bcount) ||
            bcount != sample_count * num_channels * 2) {
                fprintf (stderr, "can't write data, disk probably full!\n");
                free (sample_buffer);
                return HARD_ERROR;
        }

        samples_done += sample_count;

        if (progress != floor ((double) samples_done / num_samples * 100.0 + 0.5)) {
            int nobs = progress == -1.0;

            progress = floor ((double) samples_done / num_samples * 100.0 + 0.5);

            if (!quiet_mode)
                fprintf (stderr, "%s%3d%% done...",
                    nobs ? " " : "\b\b\b\b\b\b\b\b\b\b\b\b", (int) progress);
        }
    }

    if (!quiet_mode)
        fprintf (stderr, "\n");

    if (statistics) {
	double full_scale_rms = sqrt ((double) MAX_SAMPLE * (MAX_SAMPLE - 1.0) * 0.5);
        char temp [32];
        int chan, i;

        fprintf (stderr, "\n");
        fprintf (stderr, "%16s:%12d\n", "num channels", num_channels);
        fprintf (stderr, "%16s:%12d\n", "sample rate", 44100);
        fprintf (stderr, "%16s:%12.3f\n", "peak limit", peak_limit);
        string_time (temp, (double) num_samples / 44100);
        fprintf (stderr, "%16s:%12s\n", "duration", temp);
        fprintf (stderr, "\n");

        if (num_channels > 1) {
            fprintf (stderr, "%16s:", "channels");
            for (chan = 0; chan < num_channels; ++chan)
                fprintf (stderr, "       %2d   ", chan + 1);
            fprintf (stderr, "\n");

            for (i = 0; i < 17 + num_channels * 12; ++i)
                fputc ('-', stderr);

            fprintf (stderr, "\n");
        }

        fprintf (stderr, "%16s:", "min in value");
        for (chan = 0; chan < num_channels; ++chan)
            fprintf (stderr, "%12d", chan_data [chan].in_min);
        fprintf (stderr, "\n");

        fprintf (stderr, "%16s:", "max in value");
        for (chan = 0; chan < num_channels; ++chan)
            fprintf (stderr, "%12d", chan_data [chan].in_max);
        fprintf (stderr, "\n");

        fprintf (stderr, "%16s:", "max rms level");
        for (chan = 0; chan < num_channels; ++chan) {
            sprintf (temp, "%.2f dB", log10 (chan_data [chan].level_max / full_scale_rms) * 20.0);
            fprintf (stderr, "%12s", temp);
	}
        fprintf (stderr, "\n");

        fprintf (stderr, "%16s:", "ave rms level");
        for (chan = 0; chan < num_channels; ++chan) {
            sprintf (temp, "%.2f dB",
		log10 (chan_data [chan].level_sum / num_samples / full_scale_rms) * 20.0);
            fprintf (stderr, "%12s", temp);
	}
        fprintf (stderr, "\n\n");

        fprintf (stderr, "%16s:", "soft area");
        for (chan = 0; chan < num_channels; ++chan)
            fprintf (stderr, "%11d%%",
		(int) floor (chan_data [chan].soft_samples * 100.0 / num_samples + 0.5));
        fprintf (stderr, "\n");

        fprintf (stderr, "%16s:", "transition area");
        for (chan = 0; chan < num_channels; ++chan)
            fprintf (stderr, "%11d%%",
		(int) floor (chan_data [chan].trans_samples * 100.0 / num_samples + 0.5));
        fprintf (stderr, "\n");

        fprintf (stderr, "%16s:", "loud area");
        for (chan = 0; chan < num_channels; ++chan)
            fprintf (stderr, "%11d%%",
		(int) floor (chan_data [chan].loud_samples * 100.0 / num_samples + 0.5));
        fprintf (stderr, "\n\n");

        fprintf (stderr, "%16s:", "min out value");
        for (chan = 0; chan < num_channels; ++chan)
            fprintf (stderr, "%12d", chan_data [chan].out_min);
        fprintf (stderr, "\n");

        fprintf (stderr, "%16s:", "max out value");
        for (chan = 0; chan < num_channels; ++chan)
            fprintf (stderr, "%12d", chan_data [chan].out_max);
        fprintf (stderr, "\n\n");
    }

    free (sample_buffer);
    return NO_ERROR;
}

static double softclip (int input, int max_in, int max_out, int knee)
{
    double max_past_knee, in_past_knee, factor;

    if (input < 0)
        return -softclip (-input, max_in, max_out, knee);

    if (input < knee)
        return (double) input;

    max_past_knee = (double) max_in - knee;
    in_past_knee = (double) input - knee;
    factor = (max_in - max_out) / (max_past_knee * max_past_knee);
    return input - (in_past_knee * in_past_knee) * factor;
}

static void string_time (char *time_string, double seconds)
{
    int minutes = (int) floor (seconds / 60.0);

    sprintf (time_string, "%3d:%05.2f", minutes, seconds - (minutes * 60.0));
}

//////////////////////////// File I/O Wrapper ////////////////////////////////

static int DoReadFile (FILE *hFile, void *lpBuffer, uint32_t nNumberOfBytesToRead, uint32_t *lpNumberOfBytesRead)
{
    uint32_t bcount;

    *lpNumberOfBytesRead = 0;

    while (nNumberOfBytesToRead) {
        bcount = (uint32_t) fread ((char *) lpBuffer + *lpNumberOfBytesRead, 1, nNumberOfBytesToRead, hFile);

        if (bcount) {
            *lpNumberOfBytesRead += bcount;
            nNumberOfBytesToRead -= bcount;
        }
        else
            break;
    }

    return !ferror (hFile);
}

static int DoWriteFile (FILE *hFile, void *lpBuffer, uint32_t nNumberOfBytesToWrite, uint32_t *lpNumberOfBytesWritten)
{
    uint32_t bcount;

    *lpNumberOfBytesWritten = 0;

    while (nNumberOfBytesToWrite) {
        bcount = (uint32_t) fwrite ((char *) lpBuffer + *lpNumberOfBytesWritten, 1, nNumberOfBytesToWrite, hFile);

        if (bcount) {
            *lpNumberOfBytesWritten += bcount;
            nNumberOfBytesToWrite -= bcount;
        }
        else
            break;
    }

    return !ferror (hFile);
}
